/* This file is part of dc_qt.

    dc_qt is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    dc_qt is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dc_qt; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "dc_settings.h"
#include "dc_parse.h"
#include "util.h"
#include "debugdlg.h"

#include <qdir.h>
#include <qtextstream.h>
#include <qstringlist.h>
#include <assert.h>
#include <qmessagebox.h>
#include <qdir.h>

using namespace std;


typedef pair<int,QString> Pair;
typedef pair<QString,QString> StrPair;

dc_settings::dc_settings()
{
	user_home_dir = getenv("HOME");
	settings_file = user_home_dir + "/.dc_qt/dc_qt.conf";
	favourites_file = user_home_dir + "/.dc_qt/dc_qt.favourites";
	settings.insert( Pair(PORT,"1412") );
	settings.insert( Pair(SLOTS,"3") );
	settings.insert( Pair(ULIMIT,"1024") );
	settings.insert( Pair(CONNECTION,"Cable") );
	settings.insert( Pair(OFFSET,"0") );
	settings.insert( Pair(FONT,"Arial,12,-1,5,50,0,0,0,0,0") );
	settings.insert( Pair(PROXYPORT,"8080") );
	settings.insert( Pair(PROXYUSERNAME,QString::null));
	settings.insert( Pair(PROXYPWD,QString::null));
}

int dc_settings::init()
{
     QDir dc_dir( user_home_dir + "/.dc_qt" );
     if( !dc_dir.exists() ) {
	  dc_dir.mkdir( user_home_dir + "/.dc_qt" );
	  return NO_SETTINGS;
     }

     if(!load_from_file()) return ERROR_LOADING_PREFS;
     if(!load_favourites_file()) return ERROR_LOADING_FAVOURITES;

     return SUCCESS;
}

void dc_settings::add_share(const QString &dir)
{
	// TODO: change so that this adds files
	// for now just add directories to the list
	
	shares.insert(dir);
}

void dc_settings::del_share(const QString &dir)
{
	shares.erase(dir);
}

void dc_settings::clear_shares()
{
	shares.clear();
}

const QString dc_settings::get_share(int i)
{
	set<QString>::iterator it = shares.begin();
	while(i--) it++;

	assert(it!=shares.end());
	return *it;
}

void dc_settings::add_favourite(const QString &name,const QString &address,const QString &password)
{
	favourites.erase(name);
	favourites.insert( pair<QString,FavData>( name, FavData(address,password) ) );
}

void dc_settings::del_favourite(const QString &name)
{
	favourites.erase(name);
}

bool dc_settings::get_favourite(int i,QString &name, QString &address, QString &password)
{
	StrMap::iterator it = favourites.begin();
	while(i--) it++;
	if(it==favourites.end()) return false;

	name = (*it).first;
	address = (*it).second.ip;
	password = (*it).second.pwd;
	return true;
}

const hub_info dc_settings::get_favourite(const QString &name)
{
	StrMap::iterator it = favourites.find(name);
	if( it == favourites.end() ) return hub_info("", "", "", "");
	else return hub_info(name, (*it).second.ip, "", "", (*it).second.pwd);
}

QString dc_settings::get_setting(int key)
{
	// Try to find the given key
	Map::iterator it = settings.find(key);
	// Return NULL if not found
	if( it == settings.end() ) return NULL;
	// Else return the value
	else if( (*it).second == "" ) return QString::null;
	else return (*it).second;
}

void dc_settings::set_setting(int key, const QString &value)
{
	// An already existing value will be overwritten
	settings.erase(key);
	settings.insert( Pair(key,value) );
}

bool dc_settings::load_from_file()
{
	// Default values goes here in case they're not in the file.
	settings.insert( Pair(PORT,"1025") );
	settings.insert( Pair(SLOTS,"3") );
	settings.insert( Pair(ULIMIT,"1024") );
	settings.insert( Pair(CONNECTION,"Cable") );
	settings.insert( Pair(OFFSET,"0") );
	settings.insert( Pair(FONT,"Arial,12,-1,5,50,0,0,0,0,0") );
	settings.insert( Pair(PROXYPORT,"8080") );
	settings.insert( Pair(ENCODING,"Latin1"));
	QFile file(settings_file);
  
	if (file.open(IO_ReadOnly))
	{
		QTextStream stream(&file);
		QString inbuf;
    
		while (!stream.eof())
		{
			inbuf = stream.readLine();
      
			if (inbuf[0] == '#')
				continue;
      
			dc_parse p(inbuf, '=');

			if( get_setting_key( p.get_str(0) ) != -1 ) 
			{
				settings.erase( get_setting_key(p.get_str(0)) );
				settings.insert( Pair( get_setting_key(p.get_str(0)), p.get_str(1) ) );
			}
			else if( p.get_str(0) == "shdir" ) 
				shares.insert(p.get_str(1));
			else 
				debugWin->print("Syntax error in configuration file.\n");
		}
		return true;
	}
	else return false;
}

// this is currently broken.  A hub that has an '=' in it will not be
// read in correctly.  
bool dc_settings::load_favourites_file()
{
	QFile file(favourites_file);

	if( file.open(IO_ReadOnly) )
	{
		QTextStream stream(&file);
		QString inbuf;
		QString name,ip;
		QString pwd = QString::null;

		while( !stream.eof() )
		{
			inbuf = stream.readLine();
			if( inbuf[0] == '#' ) continue;
			dc_parse p(inbuf,'=');
			if( p.get_str(0) == "name" )
				name = p.get_str(1);
			else if( p.get_str(0) == "address" ) {
				ip = p.get_str(1);
			}
			else if( p.get_str(0) == "password") {
			     pwd  = p.get_str(1);
			     if(pwd=="<void>") pwd = QString::null;
			     favourites.insert(pair<QString,FavData>(name,FavData(ip,pwd)));
			}
		}
		return true;
	}
	else return false;
}

// Very unintelligent saving function; removes the contents in the config file
// and writes the new stuff. TODO write a better one if desired by the masses.
bool dc_settings::save_to_file()
{
	QFile file(settings_file);

	if(file.open(IO_WriteOnly)) 
	{
		QTextStream stream(&file);
    
		stream << "nick=" << get_setting(NICKNAME) << "\n";
		stream << "desc=" << get_setting(DESCRIPTION) << "\n";
		stream << "conn=" << get_setting(CONNECTION) << "\n";
		stream << "slots=" << get_setting(SLOTS) << "\n";
		if( get_setting(ADDRESS) != NULL)
			stream << "address=" << get_setting(ADDRESS) << "\n";
		stream << "port=" << get_setting(PORT) << "\n";
		stream << "passive=" << get_setting(PASSIVE) << "\n";
		if( get_setting(OFFSET) != NULL)
			stream << "offset=" << get_setting(OFFSET) << "\n";
		stream << "dldir=" << get_setting(DLDIR) << "\n";
		stream << "limit=" << get_setting(ULIMIT) << "\n";
		stream << "font=" << get_setting(FONT) << "\n";
		stream << "email=" << get_setting(EMAIL) << "\n";
		stream << "proxyaddr=" << get_setting(PROXYADDRESS) << "\n";
		stream << "proxyport=" << get_setting(PROXYPORT) << "\n";
		stream << "proxyuser=" << get_setting(PROXYUSERNAME) << "\n";
		stream << "proxypwd=" << get_setting(PROXYPWD) << "\n";
		stream << "useproxy=" << get_setting(USEPROXY) << "\n";
		stream << "usesocks=" << get_setting(USESOCKS) << "\n";
		stream << "encoding=" << get_setting(ENCODING) << "\n";
		for(int i=0;i < num_shares();i++)
			stream << "shdir=" << get_share(i) << "\n";
		
		return true;
	}

	return false;
}

bool dc_settings::save_favourites_file()
{
  QFile file(favourites_file);
  if( file.open(IO_WriteOnly) )
    {
      QTextStream stream(&file);
      QString name,ip,pwd;
      stream << "# dc_qt favourite file. Generated automatically, DO NOT EDIT.\n";

      for(int i=0;i < num_favourites();i++)
	{
	  get_favourite(i,name,ip,pwd);
	  if(pwd.isEmpty()) pwd = "<void>";
	  stream << "name=" << name << "\naddress=" << ip << "\npassword=" << pwd << "\n";
	} 

      return true;
    }
  return false;
}

int dc_settings::get_setting_key(const QString &s) const
{
	if (s == "nick") return NICKNAME;
	if (s == "desc") return DESCRIPTION;
	if (s == "conn") return CONNECTION;
	if (s == "slots") return SLOTS;
	if (s == "address") return ADDRESS;
	if (s == "port") return PORT;
	if (s == "passive") return PASSIVE;
	if (s == "limit") return ULIMIT;
	if (s == "offset") return OFFSET;
	if (s == "dldir") return DLDIR;
	if (s == "font") return FONT;
	if (s == "email") return EMAIL;
	if (s == "proxyport") return PROXYPORT;
	if (s == "proxyaddr") return PROXYADDRESS;
	if (s == "proxyuser") return PROXYUSERNAME;
	if (s == "proxypwd") return PROXYPWD;
	if (s == "useproxy") return USEPROXY;
	if (s == "usesocks") return USESOCKS;
	if (s == "encoding") return ENCODING;
	return  -1;
}

QStringList dc_settings::get_connection_types()
{
	QStringList strlst("LAN(T3)");
	strlst << "LAN(T1)";
	strlst << "Cable";
	strlst << "DSL";
	strlst << "ISDN";
	strlst << "Sattelite";
	strlst << "56Kbps";
	strlst << "33.6Kbps";
	strlst << "28.8Kbps";
	return strlst;
}

#ifdef _DEBUG
void dc_settings::print()
{
  Map::iterator mi;
  for(mi = settings.begin();mi!=settings.end();++mi) {
       debugWin->print("key: " + (*mi).first);
    if((*mi).second == "" ) debugWin->print(" value: QString::null\n");
    else debugWin->print(" value: " + (*mi).second + "\n");
  }
  
  debugWin->print("Shared directories / files:\n");
  
  set<QString>::iterator si;
  for(si = shares.begin();si != shares.end(); ++si)
       debugWin->print(*si + "\n");
}
#endif
